"""
Utilities for TCOD consoles.
"""

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal # type: ignore
from tcod.console import Console

def get_console_order(console: Console) -> Literal['C', 'F']:
    # pylint: disable=protected-access
    return console._order



"""
This is the internal event system including hooks for the context.
"""

from typing import Union, Optional, Callable, Iterator, List, Tuple, BinaryIO
from tcod.event import KeySym, Scancode, MouseButton, KeyDown, KeyUp, TextInput, Quit, \
    WindowResized, MouseMotion, MouseWheel, MouseButtonUp, MouseButtonDown, WindowEvent, \
    KMOD_NONE, KMOD_SHIFT
from ._logging import logger
from ._platform import Platform
from . import _ansi

_catchup_read_timeout = 100

TerminalEvent = Union[
    KeyDown,
    KeyUp,
    TextInput,
    Quit,
    WindowResized,
    MouseMotion,
    MouseWheel,
    MouseButtonUp,
    MouseButtonDown,
    WindowEvent,
]

class EventsManager:
    def __init__(
        self,
        platform: Platform,
        out_file: BinaryIO,
        resize_callback: Callable[[Tuple[int, int]], None]
    ) -> None:
        self._platform = platform
        self._out_file = out_file
        self._got_quit = False
        self._got_resize = False
        self._waiting_events: List[TerminalEvent] = []
        self._resize_callback = resize_callback
        self._last_mouse_motion: Optional[Tuple[int, int]] = None
        self._current_mouse_button_down: Optional[int] = None
        self._last_term_dim: Optional[Tuple[int, int]] = None
        platform.watch_quit(self._on_quit)
        platform.watch_resize(self._on_resize)
        self._catchup()

    def wait(self, timeout: Optional[float] = None) -> Iterator[TerminalEvent]:
        self._catchup()
        if self._waiting_events:
            yield from self._waiting_events
            self._waiting_events.clear()
        yield from self._handle_input(timeout)

    def get_terminal_dim(self) -> Optional[Tuple[int, int]]:
        self._got_resize = True
        self._catchup()
        return self._last_term_dim

    def _catchup(self) -> None:
        if self._got_resize:
            _ansi.save_cursor_pos(self._out_file)
            _ansi.request_get_terminal_dim(self._out_file)
            self._out_file.flush()
            while self._got_resize:
                self._waiting_events += self._handle_input(_catchup_read_timeout)

    def _on_quit(self) -> None:
        self._got_quit = True

    def _on_resize(self) -> None:
        self._got_resize = True

    def _handle_input(self, timeout: Optional[float]) -> Iterator[TerminalEvent]:
        key = self._platform.getch(timeout)
        if self._got_quit:
            yield Quit()
            self._got_quit = False
        if key is not None:
            if key == _ansi.escape:
                result = _ansi.get_escape_input(self._platform)
                if result is not None:
                    yield from self._handle_escape_input(result)
            else:
                yield from self._handle_key_press(key)

    def _handle_resize(self, event: _ansi.WindowResizeInput) -> Iterator[TerminalEvent]:
        self._got_resize = False
        _ansi.restore_cursor_pos(self._out_file)
        if event.dim != self._last_term_dim:
            self._resize_callback(event.dim)
            self._last_term_dim = event.dim
            yield WindowResized(
                type='WINDOWRESIZED',
                width=event.dim[0],
                height=event.dim[1],
            )

    def _handle_key_press(self, key: bytes) -> Iterator[TerminalEvent]:
        try:
            key_text = key.decode('ascii')
        except UnicodeDecodeError:
            return
        if key.isupper():
            key = key.lower()
            mod = KMOD_SHIFT
        else:
            mod = KMOD_NONE
        key_sym = ord(key)
        yield KeyDown(sym=key_sym, scancode=Scancode.UNKNOWN, mod=mod)
        yield TextInput(text=key_text)
        yield KeyUp(sym=key_sym, scancode=Scancode.UNKNOWN, mod=mod)

    def _handle_special_key(self, key_sym: KeySym) -> Iterator[TerminalEvent]:
        yield KeyDown(sym=key_sym, scancode=Scancode.UNKNOWN, mod=KMOD_NONE)
        yield KeyUp(sym=key_sym, scancode=Scancode.UNKNOWN, mod=KMOD_NONE)

    def _handle_mouse_motion(self, event: _ansi.MouseMotionInput) -> Iterator[TerminalEvent]:
        if self._last_mouse_motion is not None:
            motion = (
                event.pos[0] - self._last_mouse_motion[0],
                event.pos[1] - self._last_mouse_motion[1],
            )
        else:
            motion = (0, 0)
        self._last_mouse_motion = event.pos
        yield MouseMotion(position=event.pos, motion=motion, tile=event.pos)

    def _handle_mouse_button(self, event: _ansi.MouseButtonInput) -> Iterator[TerminalEvent]:
        if self._last_mouse_motion is None:
            logger.warning("mouse button but don't have position")
            return
        if event.button == 3:
            if self._current_mouse_button_down is None:
                logger.warning("mouse button up but didn't know it was down")
                return
            button = self._current_mouse_button_down
            self._current_mouse_button_down = None
            yield MouseButtonUp(
                pixel=self._last_mouse_motion,
                tile=self._last_mouse_motion,
                button=button,
            )
        else:
            if event.button == 0:
                button = MouseButton.LEFT
            elif event.button == 1:
                button = MouseButton.MIDDLE
            elif event.button == 2:
                button = MouseButton.RIGHT
            else:
                logger.warning("unhandled mouse button: %r", event)
            self._current_mouse_button_down = button
            yield MouseButtonDown(
                pixel=self._last_mouse_motion,
                tile=self._last_mouse_motion,
                button=button,
            )

    def _handle_mouse_wheel(self, event: _ansi.MouseWheelInput) -> Iterator[TerminalEvent]:
        if event.button == 0:
            yield MouseWheel(x=0, y=1, flipped=False)
        elif event.button == 1:
            yield MouseWheel(x=0, y=-1, flipped=False)
        else:
            logger.warning("unhandled mouse wheel button: %r", event)

    def _handle_escape_input(self, event: _ansi.EscapeInputEvent) -> Iterator[TerminalEvent]:
        if isinstance(event, _ansi.WindowResizeInput):
            yield from self._handle_resize(event)
        elif isinstance(event, _ansi.SpecialKeyInput):
            yield from self._handle_special_key(event.key_sym)
        elif isinstance(event, _ansi.MouseMotionInput):
            yield from self._handle_mouse_motion(event)
        elif isinstance(event, _ansi.MouseButtonInput):
            yield from self._handle_mouse_button(event)
        elif isinstance(event, _ansi.MouseWheelInput):
            yield from self._handle_mouse_wheel(event)
        elif isinstance(event, _ansi.WindowFocusGained):
            yield WindowEvent(type='WindowFocusGained')
        elif isinstance(event, _ansi.WindowFocusLost):
            yield WindowEvent(type='WindowFocusLost')
        else:
            logger.warning("unhandled escape input: %r", event)



"""
Windows platform support.

TODO: This is totally untested, and doesn't support some things!
"""

from typing import Callable, Optional, BinaryIO
import msvcrt # pylint: disable=import-error

class WindowsPlatform:
    def __init__(self, in_file: BinaryIO):
        self.in_file = in_file

    def open(self) -> None:
        pass

    def close(self) -> None:
        pass

    def getch(self, timeout: Optional[float] = None) -> Optional[bytes]:
        # pylint: disable=unused-argument
        return msvcrt.getch() # type: ignore

    def watch_resize(self, callback: Callable[[], None]) -> None:
        pass

    def watch_quit(self, callback: Callable[[], None]) -> None:
        pass



"""
ANSI terminal control.
"""

from typing import Union, Optional, Tuple, BinaryIO, NamedTuple
import dataclasses
from tcod.event import KeySym
from ._logging import logger
from ._platform import Platform

escape = b"\x1B"

class _EscapeInputResult(NamedTuple):
    start: bytes
    end: Optional[bytes]
    arg0: int
    arg1: Optional[int]

@dataclasses.dataclass(frozen=True)
class WindowResizeInput:
    dim: Tuple[int, int]

@dataclasses.dataclass(frozen=True)
class SpecialKeyInput:
    key_sym: KeySym

@dataclasses.dataclass(frozen=True)
class MouseMotionInput:
    pos: Tuple[int, int]

@dataclasses.dataclass(frozen=True)
class MouseButtonInput:
    button: int

@dataclasses.dataclass(frozen=True)
class MouseWheelInput:
    button: int

@dataclasses.dataclass(frozen=True)
class WindowFocusGained:
    pass

@dataclasses.dataclass(frozen=True)
class WindowFocusLost:
    pass

EscapeInputEvent = Union[
    WindowResizeInput,
    SpecialKeyInput,
    MouseMotionInput,
    MouseButtonInput,
    MouseWheelInput,
    WindowFocusGained,
    WindowFocusLost,
]

def reset(out_file: BinaryIO) -> None:
    out_file.write(b"%sc" % (escape))

def hide_cursor(out_file: BinaryIO) -> None:
    out_file.write(b"%s[?25l" % (escape))

def show_cursor(out_file: BinaryIO) -> None:
    out_file.write(b"%s[?25h" % (escape))

def enable_mouse_tracking(out_file: BinaryIO) -> None:
    out_file.write(b"%s[?1003h" % (escape))

def disable_mouse_tracking(out_file: BinaryIO) -> None:
    out_file.write(b"%s[?1003l" % (escape))

def enable_focus_reporting(out_file: BinaryIO) -> None:
    out_file.write(b"%s[?1004h" % (escape))

def disable_focus_reporting(out_file: BinaryIO) -> None:
    out_file.write(b"%s[?1004l" % (escape))

def request_terminal_chars_dim(dim: Tuple[int, int], out_file: BinaryIO) -> None:
    w, h = dim
    out_file.write(b"%s[8;%i;%it" % (escape, h, w))

def request_terminal_pixels_dim(dim: Tuple[int, int], out_file: BinaryIO) -> None:
    w, h = dim
    out_file.write(b"%s[4;%i;%it" % (escape, h, w))

def request_terminal_window_pos(pos: Tuple[int, int], out_file: BinaryIO) -> None:
    x, y = pos
    out_file.write(b"%s[3;%i;%it" % (escape, x, y))

def request_terminal_title(title: str, out_file: BinaryIO) -> None:
    out_file.write(b"%s]0;%s\007" % (escape, title.encode('utf8')))

def clear_screen(out_file: BinaryIO) -> None:
    out_file.write(b"%s[2J" % (escape))

def set_cursor_pos(pos: Tuple[int, int], out_file: BinaryIO) -> None:
    x, y = pos
    out_file.write(b"%s[%i;%iH" % (escape, y, x))

def save_cursor_pos(out_file: BinaryIO) -> None:
    out_file.write(b"%ss" % (escape))

def restore_cursor_pos(out_file: BinaryIO) -> None:
    out_file.write(b"%su" % (escape))

def request_get_cursor_pos(out_file: BinaryIO) -> None:
    out_file.write(b"%s[6n" % (escape))

def request_get_terminal_dim(out_file: BinaryIO) -> None:
    b = 2**15 - 1
    set_cursor_pos((b, b), out_file)
    request_get_cursor_pos(out_file)

def _read_terminated_int(
    platform: Platform,
    timeout: Optional[int],
    max_len: int = 16,
) -> Tuple[int, Optional[bytes]]:
    num = 0
    for _ in range(max_len):
        ch = platform.getch(timeout)
        if ch is None:
            break
        if not ch.isdigit():
            return num, ch
        num = num * 10 + int(ch)
    return num, None

def _read_escape_input(platform: Platform, timeout: Optional[int]) -> Optional[_EscapeInputResult]:
    start = platform.getch(timeout)
    if start not in (b'[', b'O'):
        return None
    arg0, end = _read_terminated_int(platform, timeout)
    if end == b';':
        arg1, end = _read_terminated_int(platform, timeout)
    else:
        arg1 = None
    return _EscapeInputResult(start=start, end=end, arg0=arg0, arg1=arg1)

def _get_mouse_input(platform: Platform, timeout: Optional[int]) -> Optional[EscapeInputEvent]:
    cb_ch = platform.getch(timeout)
    x_ch = platform.getch(timeout)
    y_ch = platform.getch(timeout)
    if cb_ch is None or x_ch is None or y_ch is None:
        return None
    cb = cb_ch[0]
    x = x_ch[0] - 33
    y = y_ch[0] - 33
    if cb & 32 != 0:
        if cb & 64 != 0:
            return MouseWheelInput(button=cb & 3)
        return MouseButtonInput(button=cb & 3)
    return MouseMotionInput(pos=(x, y))

def get_escape_input(
    platform: Platform,
    timeout: Optional[int] = None,
) -> Optional[EscapeInputEvent]:
    # pylint: disable=too-many-branches,too-many-return-statements
    result = _read_escape_input(platform, timeout)
    if result is None:
        return None
    if result.start == b'[': # CSI
        if result.end == b'R' and result.arg1 is not None:
            return WindowResizeInput(dim=(result.arg1, result.arg0))
        if result.end == b'M':
            return _get_mouse_input(platform, timeout)
        if result.end == b'I':
            return WindowFocusGained()
        if result.end == b'O':
            return WindowFocusLost()
        if result.end == b'A':
            return SpecialKeyInput(KeySym.UP)
        if result.end == b'B':
            return SpecialKeyInput(KeySym.DOWN)
        if result.end == b'C':
            return SpecialKeyInput(KeySym.RIGHT)
        if result.end == b'D':
            return SpecialKeyInput(KeySym.LEFT)
        if result.end == b'H':
            return SpecialKeyInput(KeySym.HOME)
        if result.end == b'F':
            return SpecialKeyInput(KeySym.END)
        if result.end == b'P':
            return SpecialKeyInput(KeySym.F1)
        if result.end == b'Q':
            return SpecialKeyInput(KeySym.F2)
        if result.end == b'R':
            return SpecialKeyInput(KeySym.F3)
        if result.end == b'S':
            return SpecialKeyInput(KeySym.F4)
        if result.end == b'~' and result.arg0 is not None:
            if result.arg0 == 1:
                return SpecialKeyInput(KeySym.HOME)
            if result.arg0 == 2:
                return SpecialKeyInput(KeySym.INSERT)
            if result.arg0 == 3:
                return SpecialKeyInput(KeySym.DELETE)
            if result.arg0 == 4:
                return SpecialKeyInput(KeySym.END)
            if result.arg0 == 5:
                return SpecialKeyInput(KeySym.PAGEUP)
            if result.arg0 == 6:
                return SpecialKeyInput(KeySym.PAGEDOWN)
            if result.arg0 == 7:
                return SpecialKeyInput(KeySym.HOME)
            if result.arg0 == 8:
                return SpecialKeyInput(KeySym.END)
            if result.arg0 == 11:
                return SpecialKeyInput(KeySym.F1)
            if result.arg0 == 12:
                return SpecialKeyInput(KeySym.F2)
            if result.arg0 == 13:
                return SpecialKeyInput(KeySym.F3)
            if result.arg0 == 14:
                return SpecialKeyInput(KeySym.F4)
            if result.arg0 == 15:
                return SpecialKeyInput(KeySym.F5)
            if result.arg0 == 17:
                return SpecialKeyInput(KeySym.F6)
            if result.arg0 == 18:
                return SpecialKeyInput(KeySym.F7)
            if result.arg0 == 19:
                return SpecialKeyInput(KeySym.F8)
            if result.arg0 == 20:
                return SpecialKeyInput(KeySym.F9)
            if result.arg0 == 21:
                return SpecialKeyInput(KeySym.F10)
            if result.arg0 == 23:
                return SpecialKeyInput(KeySym.F11)
            if result.arg0 == 24:
                return SpecialKeyInput(KeySym.F12)
    elif result.start == b'O': # SS3
        if result.end == b'P':
            return SpecialKeyInput(KeySym.F1)
        if result.end == b'Q':
            return SpecialKeyInput(KeySym.F2)
        if result.end == b'R':
            return SpecialKeyInput(KeySym.F3)
        if result.end == b'S':
            return SpecialKeyInput(KeySym.F4)
    logger.debug("unknown escape: %r", result)
    return None

def make_set_colours_true(
    fg: Tuple[int, int, int, int],
    bg: Tuple[int, int, int, int],
) -> bytes:
    return b"%s[38;2;%i;%i;%im%s[48;2;%i;%i;%im" \
        % (escape, fg[0], fg[1], fg[2], escape, bg[0], bg[1], bg[2])



"""
Abstract context type to cover both regular TCOD and terminals.
"""

from typing import TypeVar, Type, Optional, Tuple
from types import TracebackType
try:
    from typing import Literal, Protocol # pylint: disable=ungrouped-imports
except ImportError:
    from typing_extensions import Literal, Protocol # type: ignore
from tcod.console import Console
from tcod.event import Event

C = TypeVar('C', bound="TerminalCompatibleContext")
E = TypeVar("E", bound=Event)

class TerminalCompatibleContext(Protocol):
    """
    Protocol for contexts which covers basic functionality that can be supported on terminals.

    Has slightly limited features compared to a regular TCOD context.

    A regular `tcod.context.Context` should satisfy this protocol, as does `TerminalContext`.
    """

    def __enter__(self: C) -> "C":
        ...

    def __exit__(
        self,
        type_: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> None:
        ...

    def close(self) -> None:
        ...

    def present(
        self,
        console: Console,
        *,
        clear_color: Tuple[int, int, int] = (0, 0, 0),
        align: Tuple[float, float] = (0.5, 0.5)
    ) -> None:
        ...

    def pixel_to_tile(self, x: int, y: int) -> Tuple[int, int]:
        ...

    def pixel_to_subtile(self, x: int, y: int) -> Tuple[float, float]:
        ...

    def convert_event(self, event: E) -> E:
        ...

    def new_console(self, *, order: Literal['C', 'F'] = 'C') -> Console:
        ...

    def recommended_console_size(self) -> Tuple[int, int]:
        ...



"""
TCOD-compatible contexts that will write to a terminal.
"""

from typing import Optional
import sys
import os
from ._abstract_context import TerminalCompatibleContext
from ._internal_context import TerminalContext, make_terminal_context
from ._presenters import Presenter, NaivePresenter, SparsePresenter

__all__ = (
    'TerminalCompatibleContext',
    'TerminalContext',
    'new',
    'Presenter',
    'NaivePresenter',
    'SparsePresenter',
)

def new(
    *,
    x: Optional[int] = None,
    y: Optional[int] = None,
    width: Optional[int] = None,
    height: Optional[int] = None,
    columns: Optional[int] = None,
    rows: Optional[int] = None,
    title: Optional[str] = None
) -> TerminalContext:
    """
    Corresponds to `tcod.context.new()` but produces a terminal context.

    The position and dimension settings `x`, `y`, `width`, `height`, `columns`
    and `rows` are requests, which may or may not be granted; the caller should
    use the returned context's `recommended_console_size()` or `new_console()`
    to get the actual dimensions.

    This does not read `sys.argv` or take `argv` as input.
    """
    in_file = sys.stdin.buffer
    out_file = os.fdopen(sys.stdout.fileno(), 'wb', 1024)
    return make_terminal_context(
        in_file=in_file,
        out_file=out_file,
        requested_window_pos=(x, y) if x is not None and y is not None else None,
        requested_pixels_dim=(width, height) if width is not None and height is not None else None,
        requested_chars_dim=(columns, rows) if columns is not None and rows is not None else None,
        title=title,
    )



"""
This is the internal context system.
"""

from typing import TypeVar, Any, Optional, Sequence, Tuple, List, BinaryIO
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal # type: ignore
from tcod.console import Console
from tcod.event import Event
from ._platform import Platform, make_platform
from ._internal_event import EventsManager
from ._abstract_context import TerminalCompatibleContext
from ._presenters import Presenter, NaivePresenter
from . import _ansi

E = TypeVar("E", bound=Event)

_context_stack: List["TerminalContext"] = []

class TerminalContext(TerminalCompatibleContext):
    """
    TCOD-compatible context that writes to a terminal.

    The size of the terminal can be requested but is not guaranteed, and may
    change at runtime after the context is created.
    `recommended_console_size()` is the actual size of the terminal and
    `new_console()` creates a console of that size.
    """

    _out_file: BinaryIO
    _platform: Platform
    _last_term_dim: Tuple[int, int]
    _cursor_visible: bool
    _cursor_position: Tuple[int, int]
    _events_manager: EventsManager

    def _open(
        self,
        *,
        requested_window_pos: Optional[Tuple[int, int]],
        requested_pixels_dim: Optional[Tuple[int, int]],
        requested_chars_dim: Optional[Tuple[int, int]],
        title: Optional[str],
    ) -> None:
        _ansi.hide_cursor(self._out_file)
        _ansi.set_cursor_pos((0, 0), self._out_file)
        _ansi.enable_mouse_tracking(self._out_file)
        _ansi.enable_focus_reporting(self._out_file)
        if requested_window_pos is not None:
            _ansi.request_terminal_window_pos(requested_window_pos, self._out_file)
        if requested_pixels_dim is not None:
            _ansi.request_terminal_pixels_dim(requested_pixels_dim, self._out_file)
        elif requested_chars_dim is not None:
            _ansi.request_terminal_chars_dim(requested_chars_dim, self._out_file)
        if title is not None:
            _ansi.request_terminal_title(title, self._out_file)
        self._out_file.flush()

    def __enter__(self) -> "TerminalContext":
        return self

    def close(self) -> None:
        global _context_stack
        _ansi.set_cursor_pos((0, 0), self._out_file)
        _ansi.clear_screen(self._out_file)
        _ansi.show_cursor(self._out_file)
        _ansi.disable_mouse_tracking(self._out_file)
        _ansi.disable_focus_reporting(self._out_file)
        _ansi.reset(self._out_file)
        self._out_file.flush()
        self._platform.close()
        _context_stack = [c for c in _context_stack if c is not self]

    def __exit__(self, *args: Any) -> None:
        self.close()

    def present(
        self,
        console: Console,
        *,
        clear_color: Tuple[int, int, int] = (0, 0, 0),
        align: Tuple[float, float] = (0.5, 0.5),
        presenter: Optional[Presenter] = None
    ) -> None:
        """
        Present a console to this context’s display.

        `presenter` is the `Presenter` to use to control how the console is
        written to the terminal. In general the presenter instance should be
        reused between calls to `present()`. Other arguments are as for regular
        TCOD `present()`.
        """
        # pylint: disable=arguments-differ
        if presenter is None:
            presenter = NaivePresenter()
        presenter.present(
            console=console,
            term_dim=self._last_term_dim,
            align=align,
            clear_colour=clear_color,
            out_file=self._out_file
        )
        cur_x, cur_y = self._cursor_position
        _ansi.set_cursor_pos((cur_x + 1, cur_y + 1), self._out_file)
        self._out_file.flush()

    def pixel_to_tile(self, x: int, y: int) -> Tuple[int, int]:
        return x, y

    def pixel_to_subtile(self, x: int, y: int) -> Tuple[float, float]:
        return x, y

    def convert_event(self, event: E) -> E:
        return event

    def new_console(self, *, order: Literal['C', 'F'] = 'C') -> Console:
        """
        Return a new console sized for this context, that is with the actual
        size of the terminal.
        """
        width, height = self.recommended_console_size()
        return Console(width, height, order=order)

    def recommended_console_size(self) -> Tuple[int, int]:
        """
        Return the recommended size of a console for this context, which is
        always the actual size of the terminal.
        """
        new_term_dim = self._events_manager.get_terminal_dim()
        if new_term_dim is None:
            return (0, 0)
        self._last_term_dim = new_term_dim[0], new_term_dim[1]
        return self._last_term_dim

    def _on_resize(self, dim: Tuple[int, int]) -> None:
        self._last_term_dim = dim

    @property
    def cursor_position(self) -> Tuple[int, int]:
        """
        The current position of the terminal cursor.
        """
        return self._cursor_position

    @cursor_position.setter
    def cursor_position(self, value: Tuple[int, int]) -> None:
        self._cursor_position = value

    @property
    def cursor_visible(self) -> bool:
        return self._cursor_visible

    @cursor_visible.setter
    def cursor_visible(self, value: bool) -> None:
        if value:
            _ansi.show_cursor(self._out_file)
        else:
            _ansi.hide_cursor(self._out_file)
        self._cursor_visible = value

def get_terminal_context_stack() -> Sequence[TerminalContext]:
    return _context_stack

def get_events_manager(context: TerminalContext) -> EventsManager:
    # pylint: disable=protected-access
    return context._events_manager

def make_terminal_context(
    *,
    in_file: BinaryIO,
    out_file: BinaryIO,
    requested_window_pos: Optional[Tuple[int, int]] = None,
    requested_pixels_dim: Optional[Tuple[int, int]] = None,
    requested_chars_dim: Optional[Tuple[int, int]] = None,
    title: Optional[str] = None
) -> TerminalContext:
    # pylint: disable=protected-access
    new: TerminalContext = TerminalContext.__new__(TerminalContext)
    new._out_file = out_file
    new._platform = make_platform(in_file)
    new._platform.open()
    new._last_term_dim = (0, 0)
    new._cursor_visible = False
    new._cursor_position = (0, 0)
    new._events_manager = EventsManager(new._platform, new._out_file, new._on_resize)
    new._open(
        requested_window_pos=requested_window_pos,
        requested_pixels_dim=requested_pixels_dim,
        requested_chars_dim=requested_chars_dim,
        title=title,
    )
    _context_stack.append(new)
    return new



"""
Presenters which handle presenting a console on a terminal.
"""

from typing import Any, Callable, Iterable, NamedTuple, Tuple, BinaryIO
try:
    from typing import Protocol # pylint: disable=ungrouped-imports
except ImportError:
    from typing_extensions import Protocol # type: ignore
from numpy.typing import NDArray
import numpy
from tcod.console import Console
from ._console_utils import get_console_order
from ._ansi import escape, make_set_colours_true

_PAD_FG = (0, 0, 0, 0)

class Presenter(Protocol):
    """
    Presenter which handles presenting a console on a terminal.
    """

    def present(
        self,
        *,
        console: Console,
        term_dim: Tuple[int, int],
        out_file: BinaryIO,
        clear_colour: Tuple[int, int, int],
        align: Tuple[float, float]
    ) -> None:
        ...

class _DrawPlan(NamedTuple):
    draw_dim: Tuple[int, int]
    pad_left: int
    pad_top: int
    buf_get: Callable[[int, int, NDArray[Any]], Any]

def _get_draw_plan(
    console: Console,
    term_dim: Tuple[int, int],
    align: Tuple[float, float]
) -> _DrawPlan:
    order = get_console_order(console)

    con_dim = console.rgba.shape
    if order == "F":
        def buf_get(x: int, y: int, buf: NDArray[Any]) -> Any:
            return buf[x, y]
    elif order == "C":
        con_dim = con_dim[1], con_dim[0]
        def buf_get(x: int, y: int, buf: NDArray[Any]) -> Any:
            return buf[y, x]
    else:
        assert False, "unknown console order"

    draw_dim = (min(con_dim[0], term_dim[0]), min(con_dim[1], term_dim[1]))
    pad_left = int((term_dim[0] - draw_dim[0]) * align[0])
    pad_top = int((term_dim[1] - draw_dim[1]) * align[1])

    return _DrawPlan(draw_dim, pad_left, pad_top, buf_get)

def _draw_naive(
    *,
    draw_dim: Tuple[int, int],
    pad_left: int,
    pad_right: int,
    pad_top: int,
    pad_bottom: int,
    pad_bg: Tuple[int, int, int, int],
    buf_get: Callable[[int, int, NDArray[Any]], Any],
    console: Console
) -> Iterable[bytes]:
    term_y = 1

    yield make_set_colours_true(_PAD_FG, pad_bg)
    for _ in range(pad_top):
        yield b"%s[%i;1H" % (escape, term_y)
        yield b"%s[2K" % (escape)
        term_y += 1

    for con_y in range(draw_dim[1]):
        yield b"%s[%i;%iH" % (escape, term_y, pad_left + 1)
        yield make_set_colours_true(_PAD_FG, pad_bg)
        yield b"%s[1K" % (escape)
        for con_x in range(draw_dim[0]):
            c, fg, bg = buf_get(con_x, con_y, console.rgba)
            yield make_set_colours_true(fg, bg)
            yield c
        if pad_right > 0:
            yield make_set_colours_true(_PAD_FG, pad_bg)
            yield b"%s[0K" % (escape)
        term_y += 1

    yield make_set_colours_true(_PAD_FG, pad_bg)
    for _ in range(pad_bottom):
        yield b"%s[%i;1H" % (escape, term_y)
        yield b"%s[2K" % (escape)
        term_y += 1

class NaivePresenter(Presenter):
    """
    Basic presenter which always writes the whole console to the terminal.
    """

    def present(
        self,
        *,
        console: Console,
        term_dim: Tuple[int, int],
        out_file: BinaryIO,
        clear_colour: Tuple[int, int, int],
        align: Tuple[float, float]
    ) -> None:
        # pylint: disable=too-many-locals

        draw_dim, pad_left, pad_top, buf_get = _get_draw_plan(console, term_dim, align)

        out_file.write(b''.join(_draw_naive(
            draw_dim=draw_dim,
            pad_left=pad_left,
            pad_top=pad_top,
            pad_right=term_dim[0] - draw_dim[0] - pad_left,
            pad_bottom=term_dim[1] - draw_dim[1] - pad_top,
            pad_bg=clear_colour + (0,),
            buf_get=buf_get,
            console=console
        )))

def _draw_sparse_changes(
    *,
    draw_dim: Tuple[int, int],
    pad_left: int,
    pad_top: int,
    buf_get: Callable[[int, int, NDArray[Any]], Any],
    to_draw: NDArray[Any],
    console: Console
) -> Iterable[bytes]:
    for con_x, con_y in numpy.ndindex(draw_dim):
        if buf_get(con_x, con_y, to_draw):
            yield b"%s[%i;%iH" % (escape, con_y + pad_top, con_x + pad_left)
            c, fg, bg = buf_get(con_x, con_y, console.rgba)
            yield make_set_colours_true(fg, bg)
            yield c

class SparsePresenter:
    """
    Presenter which finds differences between frames and only writes the changes to the terminal.

    May be faster than the naive presenter if you are usually updating only
    small parts of the console. Needs to be reused between `present()` calls.
    """

    def __init__(self) -> None:
        self._last_buffer = numpy.full(fill_value=0, shape=(0, 0))
        self._fallback = NaivePresenter()

    def present(
        self,
        *,
        console: Console,
        term_dim: Tuple[int, int],
        out_file: BinaryIO,
        clear_colour: Tuple[int, int, int],
        align: Tuple[float, float]
    ) -> None:
        # pylint: disable=too-many-locals

        if console.rgba.shape != self._last_buffer.shape:
            self._fallback.present(
                console=console,
                term_dim=term_dim,
                out_file=out_file,
                clear_colour=clear_colour,
                align=align
            )

        else:
            draw_dim, pad_left, pad_top, buf_get = _get_draw_plan(console, term_dim, align)
            diff = console.rgba != self._last_buffer
            out_file.write(b''.join(_draw_sparse_changes(
                draw_dim=draw_dim,
                pad_left=pad_left + 1,
                pad_top=pad_top + 1,
                buf_get=buf_get,
                to_draw=diff,
                console=console
            )))

        self._last_buffer = numpy.copy(console.rgba)



"""
Protocol for platform support.
"""

from typing import Callable, Optional
try:
    from typing import Protocol # pylint: disable=ungrouped-imports
except ImportError:
    from typing_extensions import Protocol # type: ignore

class Platform(Protocol):
    def open(self) -> None:
        ...

    def close(self) -> None:
        ...

    def getch(self, timeout: Optional[float] = None) -> Optional[bytes]:
        ...

    def watch_resize(self, callback: Callable[[], None]) -> None:
        ...

    def watch_quit(self, callback: Callable[[], None]) -> None:
        ...



"""
Implementation of event handling for terminals.

This provides a `wait()` that generates the events which can be supported on
terminals. Import constants, event types, etc. from the regular `tcod.event`
module.

Supported events are given by the union type `TerminalEvent`.

Limitations on keyboard input:
- Only the shift modifier is supported.
- Key scancodes are not supported and will always be `K_UNKNOWN`.
- Key up events will be generated immediately after key down events.
"""

from typing import Optional, Iterator
try:
    from typing import Protocol # pylint: disable=ungrouped-imports
except ImportError:
    from typing_extensions import Protocol # type: ignore
from tcod.event import Event
from ._internal_context import get_terminal_context_stack, get_events_manager
from ._internal_event import TerminalEvent

__all__ = (
    'TerminalEvent',
    'TerminalCompatibleEventWait',
    'wait',
)

class TerminalCompatibleEventWait(Protocol):
    """
    Protocol for event wait interface that can be supported on terminals.

    Regular TCOD `wait()` should satisfy this protocol, as does the `wait()`
    here.
    """
    def __call__(self, timeout: Optional[float] = None) -> Iterator[Event]:
        ...

def wait(timeout: Optional[float] = None) -> Iterator[TerminalEvent]:
    """
    Block until events exist, then return an event iterator.

    `timeout` is the maximum number of seconds to wait, as for regular TCOD
    `wait()`.
    """
    context_stack = get_terminal_context_stack()
    assert context_stack, "wait() can only be called inside a context"
    return get_events_manager(context_stack[-1]).wait(timeout)



"""
ANSI terminal support for Python TCOD.
"""

import sys
import os
import atexit
from typing import *
import tcod
from . import _platform
from . import event
from . import context

__all__ = (
    'event',
    'context',
)



"""
Unix platform support.
"""

from typing import Union, Callable, Optional, List, BinaryIO
import os
import termios
import tty
import signal
import select

class UnixPlatform:
    def __init__(self, in_file: BinaryIO):
        self.old_attrs: Optional[List[Union[int, List[Union[bytes, int]]]]] = None
        # We need an extra pipe here so that we can interrupt any getch() call
        # on a signal.
        self._pipe_r, self._pipe_w = os.pipe()
        os.set_blocking(self._pipe_w, False)
        self.in_file = in_file.fileno()
        signal.set_wakeup_fd(self._pipe_w, warn_on_full_buffer=False)

    def open(self) -> None:
        self.old_attrs = termios.tcgetattr(self.in_file)
        tty.setraw(self.in_file)

    def close(self) -> None:
        if self.old_attrs is not None:
            termios.tcsetattr(self.in_file, termios.TCSADRAIN, self.old_attrs)

    def getch(self, timeout: Optional[float] = None) -> Optional[bytes]:
        ready, _rw, _rx = select.select((self.in_file, self._pipe_r), (), (), timeout)
        if self._pipe_r in ready:
            os.read(self._pipe_r, 1)
        if self.in_file in ready:
            return os.read(self.in_file, 1)
        return None

    def watch_resize(self, callback: Callable[[], None]) -> None:
        signal.signal(signal.SIGWINCH, lambda sn, sf: callback())

    def watch_quit(self, callback: Callable[[], None]) -> None:
        signal.signal(signal.SIGTERM, lambda sn, sf: callback())
        signal.signal(signal.SIGINT, lambda sn, sf: callback())
        signal.signal(signal.SIGQUIT, lambda sn, sf: callback())
        signal.signal(signal.SIGHUP, lambda sn, sf: callback())



"""
Platform-specific functionality.
"""

from typing import BinaryIO
import sys
from ._abstract_platform import Platform

__all__ = (
    'Platform',
    'make_platform',
)

try:
    from ._windows import WindowsPlatform

    def make_platform(in_file: BinaryIO) -> Platform:
        return WindowsPlatform(in_file)

except ImportError:
    from ._unix import UnixPlatform

    def make_platform(in_file: BinaryIO) -> Platform:
        return UnixPlatform(in_file)
